#include "casm.h"
#include <stdlib.h>

#define INVALID		-0x7fffffff

static char opcode_name[][32] = {
	"0nop",
	"2mov",
	"2add",
	"2sub",
	"2imul",
	"2idiv",
	"2and",
	"2or",
	"1not",
	"2xor",
	"1int",
	"1call",
	"0ret",
	"1jmp",
	"2cmp",
	"2sar",
	"2sal",
	"1je",
	"1jae",
	"1jbe",
	"1ja",
	"1jb",
	"1jz",
	"1jnz",
	"1jne",
	"2in",
	"2out",
	"1push",
	"1pop",
	"0cli",
	"0sti",
	"2fmov",
	"2fadd",
	"2fsub",
	"2fmul",
	"2fdiv",
	"2farjmp",
	"2farcall",
	"0farret",
	"1lsp",
	"1ssp",
	"2hmxs",
	"1hmxc",
	"2hmxi",
	"1htxs",
	"1htxl",
	"2htxp",
	"1slp",
};

static int size_type[] = {
	1,
	1,
	4,
	4,
	1,
	4,
	1,
	5,
	2,
	5,
	2
};

char str[256];
label_t label[512];
int last_label = 0;
int now_address = 0;
int db_size = 0;
int scan_label = 0;
char db_data[256];

char *get_op(char *s);
args_t get_argtype2(char *s);
args_t get_argtype(char *s,int num);
char *get_argstr(char *s,int num);
char get_bytecode(char *s);
unsigned int get_labeladdr(char *s);
binary_t get_binary(char *s);

int main(int argc,char *argv[])
{
	FILE *fp0;
	FILE *fp1;
	binary_t bin;
	char code[512];

	if(argc < 2) {
		puts("usage>casm <source-file> <binary-file>");
		exit(1);
	}

	fp0 = fopen(argv[1],"rt");
	fp1 = fopen(argv[2],"wb");

	scan_label = 0;

	do {
		fgets(code,512,fp0);
		bin = get_binary(code);
	} while (!feof(fp0));

	fseek(fp0,0,SEEK_SET);

	scan_label = 1;

	do {
		fgets(code,512,fp0);
		if(code[0] == 0x0d || code[0] == 0x0a) continue;
		char *p = code;
		for(;*p==' '||*p==0x09;p++);
		if(*p == ';') continue;
		bin = get_binary(code);
		if(bin.bytes!=0)fwrite(bin.bin,sizeof(char),bin.bytes,fp1);
	} while (!feof(fp0));

	fclose(fp0);
	fclose(fp1);

	return 0;
}

char *skipspace(char *p)
{
	for (; *p == ' '; p++) { }	/* Xy[Xǂݔ΂ */
	return p;
}

int getnum(char **pp, int priority)
{
	char *p = *pp;
	int i = INVALID, j;
	p = skipspace(p);

	/* PZq */
	if (*p == '+') {
		p = skipspace(p + 1);
		i = getnum(&p, 0);
	} else if (*p == '-') {
		p = skipspace(p + 1);
		i = getnum(&p, 0);
		if (i != INVALID) {
			i = - i;
		}
	} else if (*p == '~') {
		p = skipspace(p + 1);
		i = getnum(&p, 0);
		if (i != INVALID) {
			i = ~i;
		}
	} else if (*p == '(') {	/*  */
		p = skipspace(p + 1);
		i = getnum(&p, 9);
		if (*p == ')') {
			p = skipspace(p + 1);
		} else {
			i = INVALID;
		}
	} else if ('0' <= *p && *p <= '9') { /* l */
		i = strtol(p, &p, 0);
	} else { /* G[ */
		i = INVALID;
	}

	/* 񍀉Zq */
	for (;;) {
		if (i == INVALID) {
			break;
		}
		p = skipspace(p);
		if (*p == '+' && priority > 2) {
			p = skipspace(p + 1);
			j = getnum(&p, 2);
			if (j != INVALID) {
				i += j;
			} else {
				i = INVALID;
			}
		} else if (*p == '-' && priority > 2) {
			p = skipspace(p + 1);
			j = getnum(&p, 2);
			if (j != INVALID) {
				i -= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '*' && priority > 1) {
			p = skipspace(p + 1);
			j = getnum(&p, 1);
			if (j != INVALID) {
				i *= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '/' && priority > 1) {
			p = skipspace(p + 1);
			j = getnum(&p, 1);
			if (j != INVALID && j != 0) {
				i /= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '%' && priority > 1) {
			p = skipspace(p + 1);
			j = getnum(&p, 1);
			if (j != INVALID && j != 0) {
				i %= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '<' && p[1] == '<' && priority > 3) {
			p = skipspace(p + 2);
			j = getnum(&p, 3);
			if (j != INVALID && j != 0) {
				i <<= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '>' && p[1] == '>' && priority > 3) {
			p = skipspace(p + 2);
			j = getnum(&p, 3);
			if (j != INVALID && j != 0) {
				i >>= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '&' && priority > 4) {
			p = skipspace(p + 1);
			j = getnum(&p, 4);
			if (j != INVALID) {
				i &= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '^' && priority > 5) {
			p = skipspace(p + 1);
			j = getnum(&p, 5);
			if (j != INVALID) {
				i ^= j;
			} else {
				i = INVALID;
			}
		} else if (*p == '|' && priority > 6) {
			p = skipspace(p + 1);
			j = getnum(&p, 6);
			if (j != INVALID) {
				i |= j;
			} else {
				i = INVALID;
			}
		} else {
			break;
		}
	}
	p = skipspace(p);
	*pp = p;
	return i;
}

char *get_op(char *s)
{
	int i;
	char *p = s;

	for(;*p==' '||*p==0x09;p++);
	for(i = 0;*p!='(';p++,i++) {
		str[i]=*p;
	}

	str[i]=0;

	return str;
}

char *get_argstr(char *s,int num)
{
	int i;
	char *p = s;

	for(;*p==' '||*p==0x09;p++);
	for(i = 0;*p!='(';p++,i++);

	p++;

	for(i=0;i<num;i++) {
		for(;*p!=',';p++);
		p++;
	}

	for(i=0;*p!=','&&*p!=')';p++,i++) {
		str[i]=*p;
	}

	str[i]=0;

	return str;
}

char *get_argstr2(char *s)
{
	int i;
	char *p = s;

	for(;*p==' '||*p==0x09;p++);
	for(i = 0;*p!='(';p++,i++);

	p++;

	for(i=0;*p!=')';p++,i++) {
		str[i]=*p;
	}

	str[i]=0;

	return str;
}

args_t get_argtype2(char *s)
{
	int i;
	args_t args;
	char strs[128];
	char rs[16];
	char ms[16],as[16];
	int m, a;
	int rsn;
	char *n;

	char *d = s;
	char t;

	switch(*d) {
	case '%':
		d++;
		t = *d;
		if(t == 'r') args.type = ARGS_IR;
		if(t == 's') args.type = ARGS_SR;
		d++;
		for(i=0;*d!=0;d++,i++) strs[i]=*d;
		strs[i]=0;
		args.num = atoi(strs);
		if(args.num >= 32) {
			puts("error: overrun reg num");
			exit(-1);
		}
		break;
	case '@':
		d++;
		d++;
		for(i = 0; *d != ':'; d++, i++) rs[i] = *d;
		rs[i] = 0;
		d++;
		for(i = 0; *d != ']'; d++, i++) as[i] = *d;
		as[i] = 0;
		args.type = ARGS_RAE;
		if(as[0] == '%') args.type = ARGS_RAR;
		n = as;
		if(as[0] == '%') {
			n++;
			n++;
			for(i=0;*n!=0;n++,i++) strs[i]=*n;
			strs[i]=0;
			args.num = atoi(strs);
		} else a = getnum(&n,9);
		rsn = atoi(rs+2);
		if(rsn >= 32) {
			puts("error: overrun reg num");
			exit(-1);
		}
		args.num = ((rsn & 0x1f) << 32) | (a << 40);
		break;
	case '$':
		args.type = ARGS_E;
		//args.num = strtol(d+1,(char **)&strs,0);
		d++;
		args.num = getnum(&d, 9);
		break;
	case '.':
		args.type = ARGS_E;
		for(i=0;d[i]!=0 && d[i]!=',' && d[i]!=')';i++) strs[i]=d[i];
		strs[i]=0;
		args.num = get_labeladdr(strs);
		break;
	default:
		puts("error: unknown args type on memory access");
		exit(-1);
	}

	return args;
}

args_t get_argtype(char *s,int num)
{
	args_t args;
	args_t arg2;
	int i;
	char *d = get_argstr(s,num);
	char strs[128];
	char rs[16];
	char ms[16],as[16];
	char t;
	int m, a;
	int rsn;
	char *n;
	gp3200_float f;

	switch(*d) {
	case '%':
		d++;
		t = *d;
		if(t == 'r') args.type = ARGS_IR;
		if(t == 'f') args.type = ARGS_FR;
		if(t == 's') args.type = ARGS_SR;
		d++;
		for(i=0;d[i]!=0;i++) strs[i]=d[i];
		strs[i]=0;
		d = strs;
		args.num = atoi(d);
		if(args.num >= 32) {
			puts("error: overrun reg num");
			exit(-1);
		}
		args.size = 4;
		break;
	case 'b':
		d++;
		if(*d != '@') d++;
		arg2 = get_argtype2(d);
		if(arg2.type == ARGS_E) args.type = ARGS_ME;
		if(arg2.type == ARGS_IR) args.type = ARGS_MR;
		if(arg2.type == ARGS_RAE) args.type = ARGS_MRAE;
		if(arg2.type == ARGS_RAR) args.type = ARGS_MRAR;
		args.num = arg2.num;
		args.size = 1;
		break;
	case 'w':
		d++;
		if(*d != '@') d++;
		arg2 = get_argtype2(d);
		if(arg2.type == ARGS_E) args.type = ARGS_ME;
		if(arg2.type == ARGS_IR) args.type = ARGS_MR;
		if(arg2.type == ARGS_RAE) args.type = ARGS_MRAE;
		if(arg2.type == ARGS_RAR) args.type = ARGS_MRAR;
		args.num = arg2.num;
		args.size = 2;
		break;
	case 'd':
		d++;
		if(*d != '@') d++;
		arg2 = get_argtype2(d);
		if(arg2.type == ARGS_E) args.type = ARGS_ME;
		if(arg2.type == ARGS_IR) args.type = ARGS_MR;
		if(arg2.type == ARGS_RAE) args.type = ARGS_MRAE;
		if(arg2.type == ARGS_RAR) args.type = ARGS_MRAR;
		args.num = arg2.num;
		args.size = 4;
		break;
	case '$':
		args.type = ARGS_E;
		d++;
		//args.num = strtol(d,(char **)&strs,0);
		args.num = getnum(&d, 9);
		args.size = 4;
		break;
	case '@':
		d++;
		d++;
		for(i = 0; *d != ':'; d++, i++) rs[i] = *d;
		rs[i] = 0;
		d++;
		for(i = 0; *d != ']'; d++, i++) as[i] = *d;
		as[i] = 0;
		args.type = ARGS_RAE;
		if(as[0] == '%') args.type = ARGS_RAR;
		n = as;
		if(as[0] == '%') {
			n++;
			n++;
			for(i=0;*n!=0;n++,i++) strs[i]=*n;
			strs[i]=0;
			args.num = atoi(strs);
		} else a = getnum(&n,9);
		rsn = atoi(rs+2);
		if(rsn >= 32) {
			puts("error: overrun reg num");
			exit(-1);
		}
		args.num = ((rsn & 0x1f) << 32) | a;
		break;
	case '#':
		args.type = ARGS_FE;
		f = gp3200_f32_float2f32(atof(d+1));
		args.num = *((UINT32 *) &f);
		args.size = 4;
		break;
	case '.':
		args.type = ARGS_E;
		for(i=0;d[i]!=0 && d[i]!=',' && d[i]!=')';i++) strs[i]=d[i];
		strs[i]=0;
		args.num = get_labeladdr(strs);
		args.size = 4;
		break;
	default:
		puts("error: unknown args type");
		exit(-1);
	}

	return args;
}

char get_bytecode(char *s)
{
	int i;
	int j;
	args_t arg;
	char *d = get_op(s);
	char buf0[256];
	char *e;

	if(strncmp(d,"org",3) == 0 || strncmp(d,"ORG",3) == 0) {
		arg = get_argtype(s,0);
		now_address = arg.num;
		return 0xff;
	}

	if(strncmp(d,"equ",3) == 0 || strncmp(d,"EQU",3) == 0) {
		arg = get_argtype(s,1);
		e = get_argstr(s,0);
		for(i=0;i<256 && s[i]!=0x0d && s[i]!=0x0a && s[i]!=0;i++) {
			label[last_label].name[i]=e[i];
		}
		label[last_label].address = arg.num;
		last_label++;
		return 0x6f;
	}

	if(strncmp(d,"db",2) == 0 || strncmp(d,"DB",2) == 0) {
		arg = get_argtype(s,0);
		db_size = 1;
		db_data[0] = arg.num;
		return 0x7f;
	}

	if(strncmp(d,"text",4) == 0 || strncmp(d,"TEXT",4) == 0) {
		db_size = strlen(get_argstr2(s));
		db_size++;
		memcpy(db_data,get_argstr2(s),db_size);
		sprintf(buf0,db_data);
		strcpy(db_data,buf0);
		db_size = strlen(db_data) + 1;
		return 0x7f;
	}

	for(i=0;;i++) {
		if(*d>='A'&& *d<='Z') {
			for(j=0;j<strlen(d);j++) {
				d[j]+=0x20;
			}
		}
		if(strcmp(d,opcode_name[i]+1)==0) break;
	}

	return i;
}

unsigned int get_labeladdr(char *s)
{
	int i;
	char *d = s;

	for(i=0;i<512;i++) {
		if(strcmp(d,label[i].name)==0) break;
	}

	if(i >= 511) return 0;
	return label[i].address;
}

binary_t get_binary(char *s)
{
	int i;
	args_t args[2];
	binary_t bin;
	char type[2];

	memset(bin.bin,0,256);

	if(scan_label == 0) {
		if(*s=='.') {
			for(i=0;i<256 && s[i]!=0x0d && s[i]!=0x0a && s[i]!=0;i++) {
				label[last_label].name[i]=s[i];
			}
			label[last_label].address = now_address;
			bin.bytes=0;
			last_label++;
			return bin;
		}
	} else {
		if(*s=='.') {
			bin.bytes=0;
			return bin;
		}
	}

	if(*s==';') {
		bin.bytes=0;
		return bin;
	}

	unsigned char opcode = get_bytecode(s);
	char argsnum = opcode_name[opcode][0]-'0';

	if(opcode==0xff) {
		bin.bytes=0;
		now_address+=bin.bytes;
		return bin;
	}

	if(opcode==0x6f) {
		bin.bytes=0;
		now_address+=bin.bytes;
		return bin;
	}

	if(db_size != 0) {
		memcpy(bin.bin,db_data,db_size);
		bin.bytes = db_size;
		db_size=0;
		now_address+=bin.bytes;
		return bin;
	}

	if(argsnum==0) {
		bin.bin[0] = opcode << 2;
		bin.bytes=1;
		now_address+=bin.bytes;
		return bin;
	}

	for(i=0;i<argsnum;i++) {
		args[i] = get_argtype(s,i);
		type[i]=args[i].type;
	}

	bin.bin[0] = (opcode << 2) | ((args[0].size - 1) & 0x03);
	if(args[0].type != ARGS_MR || args[0].type != ARGS_ME || args[0].type != ARGS_MRAE || args[0].type != ARGS_MRAR) bin.bin[0] = (opcode << 2) | ((args[1].size - 1) & 0x03);

	if(argsnum==1) {
		bin.bin[1]=type[0];
		bin.bytes=1+1+size_type[type[0]];
		*((unsigned int *)&(bin.bin[2]))=args[0].num;
	} else if(argsnum==2) {
		bin.bin[1]=(type[1] << 4)|type[0];
		bin.bytes=1+1+size_type[type[0]]+size_type[type[1]];
		*((UINT64 *)&(bin.bin[2]))=args[0].num;
		*((UINT64 *)&(bin.bin[size_type[type[0]]+2]))=args[1].num;
	}

	now_address+=bin.bytes;
	return bin;
}